#include <regex>
#include <map>

#include <iostream>
#include <vector>
#include <cmath>
#include <numeric>
#include <string>
#include <algorithm>
#include <fstream>
#include <ctime>
#include <fstream>
#include <filesystem>
#include "/usr/local/mdc_sdk_llvm/dp_gea/mdc_cross_compiler/sysroot/usr/include/opencv4/opencv2/imgcodecs.hpp"
#include "/usr/local/mdc_sdk_llvm/dp_gea/mdc_cross_compiler/sysroot/usr/include/opencv4/opencv2/imgproc.hpp"
#include "preprocess.h"
#include "model_process.h"
#include "postprocess.h"
#include "fcos3d_pipeline.h"
#include "acl/acl.h"
#include "utils.h"
#include "yaml-cpp/yaml.h"

#define CLOCK_Enable

using namespace std;
namespace fs = std::__fs::filesystem;

extern bool g_isDevice;
template<class Type>
std::shared_ptr<Type> MakeSharedNoThrow() {
    try {
        return std::make_shared<Type>();
    } catch (...) {
        return nullptr;
    }
}

#define MAKE_SHARED_NO_THROW(memory, memory_type) \
    memory = MakeSharedNoThrow<memory_type>();

Fcos3dPipeline::Fcos3dPipeline() :
        deviceId_(0), context_(nullptr), stream_(nullptr) {
}

Fcos3dPipeline::~Fcos3dPipeline() {
    DestroyResource();
}

Result Fcos3dPipeline::Init(const Fcos3dInitParams& fcos3DInitParams) {
    // ACL init
    // acl.json is dump or profiling config file
    if (!Utils::validStatus("aclInit", aclInit(nullptr))) {
        return FAILED;
    }

    if (!Utils::validStatus("aclrtSetDevice", aclrtSetDevice(deviceId_))) {
        ERROR_LOG("DeviceId is %d", deviceId_);
        (void) aclFinalize();
        return FAILED;
    }
    const char *soc_name{aclrtGetSocName()};
    if (soc_name == nullptr) {
        ERROR_LOG("Can not get the soc name");
    }
    INFO_LOG("soc name: %s", soc_name);

    // create context (set current)
    if (!Utils::validStatus("aclrtCreateContext", aclrtCreateContext(&context_, deviceId_))) {
        ERROR_LOG("deviceId = %d", deviceId_);
        return FAILED;
    }

    // create stream
    if (!Utils::validStatus("aclrtCreateStream", aclrtCreateStream(&stream_))) {
        ERROR_LOG("deviceId = %d", deviceId_);
        return FAILED;
    }

    // get run mode
    // runMode is ACL_HOST which represents app is running in host
    // runMode is ACL_DEVICE which represents app is running in device
//    aclrtRunMode runMode;
    if (!Utils::validStatus("aclrtGetRunMode", aclrtGetRunMode(&Utils::runMode_))) {
        return FAILED;
    }
    g_isDevice = (Utils::runMode_ == ACL_DEVICE);
    this->ParseConfigFile(fcos3DInitParams.config_file);

    Result ret = modelProcess.LoadModel(fcos3DConstantParameter.om_model_path.c_str());
    if (ret != SUCCESS) {
        ERROR_LOG("execute LoadModel failed");
        return FAILED;
    }

    ret = modelProcess.CreateModelDesc();
    if (ret != SUCCESS) {
        ERROR_LOG("execute CreateModelDesc failed");
        return FAILED;
    }
    // create output dataset(for inference result saving) according to modelDesc
//    ret = processModel.CreateOutput();
//    if (ret != SUCCESS) {
//        ERROR_LOG("execute CreateOutput failed");
//        return FAILED;
//    }

    return SUCCESS;
}

// Function to extract the numeric part from the filename
int extractNumber(const std::string &filename) {
    std::regex re("\\d+");
    std::smatch match;
    if (std::regex_search(filename, match, re)) {
        return std::stoi(match.str());
    }
    return -1; // Return a default value if no number is found
}

Result Fcos3dPipeline::Process() {

    size_t devBufferSize;
    void *picDevBuffer = nullptr;
//    Result ret = modelProcess.GetInputSizeByIndex(0, devBufferSize);
//    if (ret != SUCCESS) {
//        ERROR_LOG("execute GetInputSizeByIndex failed");
//        return FAILED;
//    }
//
//    aclError aclRet = aclrtMalloc(&picDevBuffer, devBufferSize, ACL_MEM_MALLOC_NORMAL_ONLY);
//    if (!Utils::validStatus("aclrtMalloc", aclRet)) {
//        ERROR_LOG("input for model size is %zu", devBufferSize);
//        return FAILED;
//    }
//
//
//    vector<DataInfo> inputData = {{picDevBuffer, devBufferSize}};
//
//
//    ret = modelProcess.CreateInput(inputData);
//    if (ret != SUCCESS) {
//        ERROR_LOG("execute CreateInput failed");
//        aclrtFree(picDevBuffer);
//        return FAILED;
//    }
//

    Result ret = modelProcess.CreateOutput();
    if (ret != SUCCESS) {
        aclrtFree(picDevBuffer);
        ERROR_LOG("execute CreateOutput failed");
        return FAILED;
    }

    // #ifdef CLOCK_Enable
    clock_t start, end, start_preprocess, end_preprocess, start_postprocess, end_postprocess;
    double total_time_preprocess = 0;
    double total_time = 0;
    double total_time_postprocess = 0;
    double total_time_procedure = 0;
    // for(size_t i=0; i<20; i++){

    FCos3dPreProcessor fCos3DPreProcessor(this->fcos3DConstantParameter);
    fCos3DPreProcessor.Init(this->fcos3DConstantParameter);

    FCos3dPostProcessor<float> fCos3DPostProcessor;
    fCos3DPostProcessor.Init(this->fcos3DConstantParameter);

    // #endif

    WARN_LOG("start to preprocess of model.");

    std::map<int, std::string> sortedFiles;

    // Read the file paths and store them in a map to sort by the numeric part of the filename
    for (const auto &entry : fs::directory_iterator(fcos3DConstantParameter.input_folder)) {
        if (entry.is_regular_file() && entry.path().extension() == ".jpg") { //.png
            std::string filePath = entry.path().string();
            int fileNumber = extractNumber(entry.path().filename().string());
            sortedFiles.insert({fileNumber, filePath});
        }
    }

    int i = 0; // Initialize the index variable    

    // Process files in the sorted order
    for (const auto &[key, filePath] : sortedFiles) {
        shared_ptr<ImageDesc> image_desc = nullptr;
        MAKE_SHARED_NO_THROW(image_desc, ImageDesc);
        if (image_desc == nullptr) {
            ERROR_LOG("Failed to MAKE_SHARED_NO_THROW for ImageDesc.");
            return FAILED;
        }

        start_preprocess = clock();
        if (!Utils::PreProcess(image_desc, filePath)) {
            continue;
        }
        Result ret = modelProcess.CreateInput((void*) image_desc->data.get(), image_desc->size);
        end_preprocess = clock();

        start = clock();
        ret = modelProcess.Execute();
        end = clock();
        if (ret != SUCCESS) {
            ERROR_LOG("execute inference failed");
            aclrtFree(picDevBuffer);
            return FAILED;
        }

        auto model_output = modelProcess.OutputModelResult();

        start_postprocess = clock();
        auto final_results = fCos3DPostProcessor.Process(model_output, fcos3DConstantParameter);
        all_results_.push_back(final_results);
        end_postprocess = clock();
        INFO_LOG("After Post Process!!!!");
        std::cout << "image path is = " << filePath << std::endl;

        std::string imageName = fs::path(filePath).stem().string(); // Get the filename without extension
        std::string outputFolderPath = fcos3DConstantParameter.output_folder;
        std::string outputFilePath_bboxes = outputFolderPath + imageName + "_bboxes" + ".txt";
        std::string outputFilePath_scores = outputFolderPath + imageName + "_scores" + ".txt";
        std::string outputFilePath_labels = outputFolderPath + imageName + "_labels" + ".txt";

        std::ofstream outputFile_bboxes(outputFilePath_bboxes);
        std::ofstream outputFile_scores(outputFilePath_scores);
        std::ofstream outputFile_labels(outputFilePath_labels);

        if (outputFile_bboxes.is_open()) {
            outputFile_bboxes << final_results.bboxes << std::endl;
            outputFile_bboxes.close();
        } else {
            std::cerr << "Failed to open output file: " << outputFilePath_bboxes << std::endl;
        }

        if (outputFile_scores.is_open()) {
            outputFile_scores << final_results.scores << std::endl;
            outputFile_scores.close();
        } else {
            std::cerr << "Failed to open output file: " << outputFilePath_scores << std::endl;
        }

        if (outputFile_labels.is_open()) {
            outputFile_labels << final_results.labels << std::endl;
            outputFile_labels.close();
        } else {
            std::cerr << "Failed to open output file: " << outputFilePath_labels << std::endl;
        }

        i++; // Increment the index variable

        #ifdef CLOCK_Enable
        double cpu_time_total = ((double) (end - start)) * 1000 / CLOCKS_PER_SEC;
        total_time += cpu_time_total;
        double cpu_time_total_preprocess = ((double) (end_preprocess - start_preprocess)) * 1000 / CLOCKS_PER_SEC;
        total_time_preprocess += cpu_time_total_preprocess;
        double cpu_time_total_postprocess = ((double) (end_postprocess - start_postprocess)) * 1000 / CLOCKS_PER_SEC;
        total_time_postprocess += cpu_time_total_postprocess;
        double cpu_time_total_procedure = ((double) (end_postprocess - start_preprocess)) * 1000 / CLOCKS_PER_SEC;
        total_time_procedure += cpu_time_total_procedure;
        
    }












//     for (const auto &entry: fs::directory_iterator(fcos3DConstantParameter.input_folder)) {
//         if (entry.is_regular_file() && entry.path().extension() == ".jpg") {
//             std::string filePath = entry.path().string();

//             // create image_desc for saving input image.
//             shared_ptr<ImageDesc> image_desc = nullptr;
//             MAKE_SHARED_NO_THROW(image_desc, ImageDesc);
//             if (image_desc == nullptr) {
//                 ERROR_LOG("Failed to MAKE_SHARED_NO_THROW for ImageDesc.");
//                 return FAILED;
//             }



// //            cv::Mat image = cv::imread(filePath); //BGR
// //            if (image.empty()) {
// //                std::cerr << "Failed to read image: " << filePath << std::endl;
// //                continue; // Skip to the next image if loading fails
// //            }
// //            cv::Mat RGBMat;
// //            cv::cvtColor(image, RGBMat, cv::COLOR_BGR2RGB);

// //            std::vector<uchar> *imageVectorPtr = fCos3DPreProcessor.matToVectorPointer(image);

// //            cv::Mat imageMat = fCos3DPreProcessor.vectorPointerToMat(imageVectorPtr,
// //                                                                     fCos3DPreProcessor.get_param().height,
// //                                                                     fCos3DPreProcessor.get_param().width);

//             start_preprocess = clock();
// //            auto output_of_preprocess =fCos3DPreProcessor.Process(imageVectorPtr);
// //            auto output_of_preprocess =fCos3DPreProcessor.image_vector_without_padding(image);
// //            auto output_of_preprocess =fCos3DPreProcessor.image_vector_without_padding_channel_last(image);
//             if (!Utils::PreProcess(image_desc, filePath)) {
//                 continue;
//             }
//             Result ret = modelProcess.CreateInput((void*) image_desc->data.get(), image_desc->size);
//             end_preprocess = clock();
// //
// ////            delete imageVectorPtr;
// //            WARN_LOG("start to prepare inputs of model from preprocess");
// //
// //            ret = Utils::MemcpyVariableToDeviceBuffer(output_of_preprocess, picDevBuffer, devBufferSize);
// //
// //            if (ret != SUCCESS) {
// //                aclrtFree(picDevBuffer);
// //                ERROR_LOG("memcpy device buffer failed, index is %d", 0);
// //                return FAILED;
// //            }
//             start = clock();
//             ret = modelProcess.Execute();
//             end = clock();
//             if (ret != SUCCESS) {
//                 ERROR_LOG("execute inference failed");
//                 aclrtFree(picDevBuffer);
//                 return FAILED;
//             }

//             auto model_output = modelProcess.OutputModelResult();

//             start_postprocess = clock();
//             auto final_results =fCos3DPostProcessor.Process(model_output,fcos3DConstantParameter);
//             final_results_ = final_results;
//             end_postprocess = clock();
//             INFO_LOG("After Post Process!!!!");

//             std::string imageName = entry.path().filename().stem(); // Get the filename without extension
//             std::string outputFolderPath = fcos3DConstantParameter.output_folder;
//             std::string outputFilePath_bboxes = outputFolderPath + imageName + "_bboxes" + ".txt";
//             std::string outputFilePath_scores = outputFolderPath + imageName + "_scores" + ".txt";
//             std::string outputFilePath_labels = outputFolderPath + imageName + "_labels" + ".txt";

//             std::ofstream outputFile_bboxes(outputFilePath_bboxes);
//             std::ofstream outputFile_scores(outputFilePath_scores);
//             std::ofstream outputFile_labels(outputFilePath_labels);

//             if (outputFile_bboxes.is_open()) {

//                 outputFile_bboxes << final_results.bboxes << std::endl;
//                 outputFile_bboxes.close();

//             } else {
//                 std::cerr << "Failed to open output file: " << outputFilePath_bboxes << std::endl;
//             }

//             if (outputFile_scores.is_open()) {

//                 outputFile_scores << final_results.scores << std::endl;
//                 outputFile_scores.close();

//             } else {
//                 std::cerr << "Failed to open output file: " << outputFilePath_scores << std::endl;
//             }

//             if (outputFile_labels.is_open()) {

//                 outputFile_labels << final_results.labels << std::endl;
//                 outputFile_labels.close();

//             } else {
//                 std::cerr << "Failed to open output file: " << outputFilePath_labels << std::endl;
//             }

//         }

// #ifdef CLOCK_Enable

//         double cpu_time_total = ((double) (end - start)) * 1000 / CLOCKS_PER_SEC;
//         total_time += cpu_time_total;
//         double cpu_time_total_preprocess = ((double) (end_preprocess - start_preprocess)) * 1000 / CLOCKS_PER_SEC;
//         total_time_preprocess += cpu_time_total_preprocess;
//         double cpu_time_total_postprocess = ((double) (end_postprocess - start_postprocess)) * 1000 / CLOCKS_PER_SEC;
//         total_time_postprocess += cpu_time_total_postprocess;
//         double cpu_time_total_procedure = ((double) (end_postprocess - start_preprocess)) * 1000 / CLOCKS_PER_SEC;
//         total_time_procedure += cpu_time_total_procedure;
//     }

    // }

    std::cout << "time of preprocess: " << total_time_preprocess / 81 << std::endl;
    std::cout << "time of execution: " << total_time / 81 << std::endl;
    std::cout << "time of postprocess: " << total_time_postprocess / 81 << std::endl;
    std::cout << "time of time of process: " << total_time_procedure / 81 << std::endl;
#endif

    modelProcess.DestroyInput();
    modelProcess.DestroyOutput();

    aclrtFree(picDevBuffer);
    return SUCCESS;
}

void Fcos3dPipeline::DestroyResource() {
    aclError ret;
    if (stream_ != nullptr) {
        ret = aclrtDestroyStream(stream_);
        if (ret != ACL_SUCCESS) {
            ERROR_LOG("destroy stream failed, errorCode = %d", static_cast<int32_t>(ret));
        }
        stream_ = nullptr;
    }
    INFO_LOG("end to destroy stream");

    if (context_ != nullptr) {
        ret = aclrtDestroyContext(context_);
        if (ret != ACL_SUCCESS) {
            ERROR_LOG("destroy context failed, errorCode = %d", static_cast<int32_t>(ret));
        }
        context_ = nullptr;
    }
    INFO_LOG("end to destroy context");

    ret = aclrtResetDevice(deviceId_);
    if (ret != ACL_SUCCESS) {
        ERROR_LOG("reset device %d failed, errorCode = %d", deviceId_, static_cast<int32_t>(ret));
    }
    INFO_LOG("end to reset device %d", deviceId_);

    ret = aclFinalize();
    if (ret != ACL_SUCCESS) {
        ERROR_LOG("finalize acl failed, errorCode = %d", static_cast<int32_t>(ret));
    }
    INFO_LOG("end to finalize acl");
}

bool Fcos3dPipeline::ParseConfigFile(const std::string &config_file) {
    // read YAML file
    const YAML::Node &config = YAML::LoadFile(config_file);
    fcos3DConstantParameter.fpn0_size = config["fpn0_size"].as<array<uint32_t, 2>>();
    fcos3DConstantParameter.fpn1_size = config["fpn1_size"].as<array<uint32_t, 2>>();
    fcos3DConstantParameter.fpn2_size = config["fpn2_size"].as<array<uint32_t, 2>>();
    fcos3DConstantParameter.fpn3_size = config["fpn3_size"].as<array<uint32_t, 2>>();
    fcos3DConstantParameter.fpn4_size = config["fpn4_size"].as<array<uint32_t, 2>>();
    fcos3DConstantParameter.strides=config["strides"].as<Vec<int>>();
    fcos3DConstantParameter.num_cls = config["num_cls"].as<uint32_t>();
    fcos3DConstantParameter.num_bbox = config["num_bbox"].as<uint32_t>();
    fcos3DConstantParameter.num_dir_cls = config["num_dir_cls"].as<uint32_t>();
    fcos3DConstantParameter.num_arrt = config["num_arrt"].as<uint32_t>();
    fcos3DConstantParameter.num_centerness = config["num_centerness"].as<uint32_t>();
    fcos3DConstantParameter.om_model_path= config["om_model_path"].as<std::string>();
    fcos3DConstantParameter.input_folder= config["input_folder"].as<std::string>();
    fcos3DConstantParameter.output_folder= config["output_folder"].as<std::string>();
    fcos3DConstantParameter.group_reg_dims= config["group_reg_dims"].as<Vec<int>>();
    fcos3DConstantParameter.origin= config["origin"].as<Vec<float>>();
    fcos3DConstantParameter.origin_defualt= config["origin_defualt"].as<Vec<float>>();
    fcos3DConstantParameter.scale_factor= config["scale_factor"].as<Vec<float>>();
    fcos3DConstantParameter.num_attrs = config["num_attrs"].as<int>();

    fcos3DConstantParameter.nms_pre = config["postprocess_params"]["nms_pre"].as<int>();
    fcos3DConstantParameter.dir_offset = config["postprocess_params"]["dir_offset"].as<float>();
    fcos3DConstantParameter.cfg_score_thr = config["postprocess_params"]["cfg_score_thr"].as<float>();
    fcos3DConstantParameter.cfg_max_per_img = config["postprocess_params"]["cfg_max_per_img"].as<int>();
    fcos3DConstantParameter.cfg_use_rotate_nms = config["postprocess_params"]["cfg_use_rotate_nms"].as<bool>();
    fcos3DConstantParameter.cfg_nms_thr = config["postprocess_params"]["cfg_nms_thr"].as<float>();
    fcos3DConstantParameter.cls_out_channels = config["postprocess_params"]["cls_out_channels"].as<int>();
    fcos3DConstantParameter.nuscenes_cam_back_cam2img = config["postprocess_params"]["nuscenes_cam_back_cam2img"].as<Vec2D<float>>();


    fcos3DConstantParameter.mean = config["preprocess_params"]["mean"].as<std::array<float, 3>>();
    fcos3DConstantParameter.std_dev = config["preprocess_params"]["std_dev"].as<std::array<float, 3>>();
    fcos3DConstantParameter.height = config["preprocess_params"]["height"].as<int>();
    fcos3DConstantParameter.width = config["preprocess_params"]["width"].as<int>();
    fcos3DConstantParameter.channel = config["preprocess_params"]["channel"].as<int>();
    fcos3DConstantParameter.num_threads = config["preprocess_params"]["num_threads"].as<int>();




    return true;
}
